/******************
input-output functions 
******************/


#include <stdio.h>
#include <sys/types.h>
#ifndef VXWORKS
#include <time.h>
#include <termios.h>
#endif
#include <math.h>
#include <unistd.h>

#include "dbliste.h"
#include "lasdef.h"
#include "las_prot.h"
#include "image_mgr_p.h"
#include "las_io_p.h"

long wrt_file_sg_c(igla *igl,const char *file)
     
{

  
  
  PDBL *l, *l_end;
  Sgl *s;
  FILE *ptr;
 

   if(igl->lsgl == NULL)
    return(-1);

    
  if((ptr = fopen(file,"w+")) == NULL){
    printf("can't open file %s\n",file);
    return(-1);
  }
  fprintf(ptr,"%ld \n",igl->nsg);
    
  fprintf(ptr,"%.8f %.8f %.8f %.8f %.8f %.8f %c \n",
	  igl->x,igl->y,igl->cs,igl->sn,igl->cst,igl->snt,igl->coosys);
  fprintf(ptr,"%.8f %.8f %.8f %.8f \n",igl->enc[0],igl->enc[1],
	  igl->enc[2],igl->enc[3]);

  l_end = dlfin(igl->lsgl);
  for(l = l_end; l != NULL; l = dlp(l)){ 

    s = (Sgl *)dlel(l);
    
    fprintf(ptr, "%.8f ",s->first.x);
    fprintf(ptr, "%.8f \n",s->first.y);
    fprintf(ptr, "%.8f ",s->last.x);
    fprintf(ptr, "%.8f \n",s->last.y);
    fprintf(ptr, "%.9f %.9f %.9f \n",s->vx,s->vy,s->d);
    fprintf(ptr, "%.9f %.9f \n",s->r,s->o);
    fprintf(ptr, "%.9f %.9f %.9f \n",s->vr,s->vro,s->vo);
    fprintf(ptr, "%c %c \n",s->typf,s->typl);
    

  }

  fclose(ptr);
  return(1);
}


long rd_file_sg_c(igla *ig, const char *file)
     
  {
    FILE *ptr,*fopen();
    long i, j=0;
    Sgl *s;
    PDBL *l;
    double vx, vy, length, dx, dy, o, d;

    
    if((ptr = fopen(file,"r")) == NULL){
      printf("Can't open %s\n",file);
      return(-1);
    }
    
    if(ig == NULL)
      return(-1);
    
    fscanf(ptr,"%ld",&(ig->nsg));
    
    ig->tbp = NULL;
    ig->nbp = 0;
    ig->lsgl = NULL;
    
    fscanf(ptr,"%lf %lf %lf %lf %lf %lf %c \n",
	   &(ig->x),&(ig->y),&(ig->cs),&(ig->sn),&(ig->cst),&(ig->snt),&(ig->coosys));
    fscanf(ptr,"%lf %lf %lf %lf \n",&ig->enc[0],&ig->enc[1],&ig->enc[2],&ig->enc[3]);
    for(i = 0; i < ig->nsg ; i++){
      s = Sgl_cree();

      fscanf(ptr, "%lf ",&(s->first.x));
      fscanf(ptr, "%lf \n",&(s->first.y));
      fscanf(ptr, "%lf ",&(s->last.x));
      fscanf(ptr, "%lf ",&(s->last.y));
      fscanf(ptr, "%lf %lf %lf \n",&(s->vx),&(s->vy),&(s->d));
      fscanf(ptr, "%lf %lf \n",&(s->r),&(s->o));
      fscanf(ptr, "%lf %lf %lf \n",&(s->vr),&(s->vro),&(s->vo));


      dx = s->last.x - s->first.x;
      dy = s->last.y - s->first.y;
      length = sqrt(dx*dx+dy*dy);
      vx = -dy/length;
      vy = dx/length;
      o = atan2(vy,vx);
      d = -(vx * s->first.x + vy * s->first.y);

      j++;
      if (fabs(vx-s->vx>0.001))
	printf ("rd_file seg %ld: vx %f -> %f\n", j, s->vx, vx);
      s->vx = vx;
      if (fabs(vy-s->vy>0.001))
	printf ("rd_file seg %ld: vy %f -> %f\n", j, s->vy, vy);
      s->vy = vy;
      if (fabs(o-s->o>0.0001))
	printf ("rd_file seg %ld: o %f -> %f\n", j, s->o, o);
      s->o = o;
      if (fabs(d-s->d>0.001))
	printf ("rd_file seg %ld: d %f -> %f\n", j, s->d, d);
      s->d = s->r = d;


      fscanf(ptr, "%c %c \n",&(s->typf),&(s->typl));
      s->first.r = sqrt(s->first.x * s->first.x + s->first.y * s->first.y);
      s->last.r = sqrt(s->last.x * s->last.x + s->last.y * s->last.y);
      ig->lsgl = dlhcons(ig->lsgl,s);
    }
    

    
    fclose(ptr);
    
    /****** assigning the possible extremity segments *****/

    for(l = ig->lsgl; l != NULL; l = dls(l)){
      s = (Sgl *)dlel(l);

      if(s->typf == 'V')
	assign_extr(s,ig->lsgl);
    }
    return(1);
    
  }
/***** look for a segment s1  in the list l such that 
  s->first = s1->last
*****/

void assign_extr(Sgl *s,PDBL *l)  
  {

    double x1,y1,x2,y2;
    PDBL *l2;
    Sgl *s2;
    
    x1 = s->first.x;
    y1 = s->first.y;

    
    for(l2 = l; l2 != NULL; l2 = dls(l2)){
      s2 = (Sgl *)dlel(l2);

      x2 = s2->last.x;
      y2 = s2->last.y;

      if( fabs( x1 -x2) < 1e-7 && fabs(y1 - y2) < 1e-7){
	s->sfirst = s2;
	s2->slast = s;
	return;
      }


    }
  
    printf("Warning: no extremity found (x1 %f y1 %f x2 %f y2 %f)\n", 
	   x1, y1, x2, y2);
    s->typf = 'B';
  }
  

#ifndef VXWORKS  
long read_char(char *c){
  static long init = 0;
  static struct termios tp1,tp2;

  if(!init){
    tcgetattr(0,&tp1);
    tcgetattr(0,&tp2);
    tp2.c_cc[VMIN] = 0;
    tp2.c_cc[VTIME] = 0;
    tp2.c_lflag &= ~ICANON;
    init = 1;
  }

  tcsetattr(0,TCSANOW,&tp2);
  if(read(0,c,1)){
    tcsetattr(0,TCSANOW,&tp1);
    return(1);
    }
  tcsetattr(0,TCSANOW,&tp1);
  return(0);

}

  
long freeze_on_rchar(char *c){
  static long init = 0;
  static struct termios tp1,tp2;

  if(!init){
    tcgetattr(0,&tp1);
    tcgetattr(0,&tp2);
    tp2.c_cc[VMIN] = 1;
    tp2.c_cc[VTIME] = 0;
    tp2.c_lflag &= ~ICANON;
    tp2.c_lflag &= ~ECHO;
    init = 1;
  }

  tcsetattr(0,TCSANOW,&tp2);
  if(read(0,c,1)){
    tcsetattr(0,TCSANOW,&tp1);
    return(1);
    }
  tcsetattr(0,TCSANOW,&tp1);
  return(0);

}
     
#endif












