


#include <stdio.h>
#include <stdlib.h>



#include "dbliste.h"





PDBL TEST_PDBL;

/********************************************************
  This file deals with double chained list structures.
  The structure itself is called a PDBL structure and has the
  following format:

  struct PDBListe { struct PDBListe *last;
                   struct PDBListe *next; 
                   void *el ; } PDBL;

  ->last points to the previous PDBL structure in the list
  ->next points to the next PDBL structure in the list.
  ->el points to the actual element included in the list.

  el can obviously be a list istself.

  A PDBList can be "closed" which means than it does not have a 
  begining or an end. Otherwise specified, functions are designed
  to handle such cases.

  WARNING: there exists a global variable called TEST_PDBL which is
  needed especially to report problem or special cases in list searching
  functions.

  NOTE: all functions use access macros defined in PDBListe.h as
  follows:

  #define dls(l) (l)->next
  #define dlp(l) (l)->last
  #define dlel(l) (l)->el
  
*****************************************************************/

/**************************************************************
  Construct functions 
  ***************************************************************/



/***************
 * FUNCTION:     dlhcons
 * PURPOSE:      adds an element at the top of the list
 * ARGUMENTS:    PDBL *dl :
 *               void *el :
 * ALGORITHM:    creates a PDBL structure, puts it right before "dl".
                 Can be done on a NULL list = starting a new list
 * RETURN:       PDBL * : pointer on the created structure
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/




PDBL *dlhcons(PDBL *dl, void *el)
{

    PDBL *l,*la;

    l = (PDBL *)malloc(sizeof(PDBL));
    l->el = el;
    la = dl;

    if(la != NULL){
	if(la->last != NULL)
	    la->last->next = l;
	l->last = la->last;
	la->last = l;
	l->next = la;
    }

    else
	l->next = l->last = NULL;

    
    return(l);
    
}

/***************
 * FUNCTION:     dlbcons
 * PURPOSE:      adds an element at the end of the list
 * ARGUMENTS:    PDBL *dl :
 *               void *el :
 * ALGORITHM:    creates a PDBL structure containing el
                 puts it right after "dl".
                 Can be done on a NULL list = starting
		 a new list.
 * RETURN:       PDBL * : pointer on the created structure
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


PDBL *dlbcons(PDBL *dl,void *el)
    
{

    PDBL *l,*la;
    
    l = (PDBL *)malloc(sizeof(PDBL));
    l->el = el;
    la = dl;

    if(la != NULL){
	if(la->next != NULL)
	    la->next->last = l;
	l->next = la->next;
	la->next = l;
	l->last = la;
    }

    else
	l->next = l->last = NULL;
    return(l);

}


/***************
 * FUNCTION:     insere_PDBL
 * PURPOSE:      inserts the PDBL pointer *PDBL right before
                 a PDBL pointer in the list *dl such that 
		 test_el(PDBL->el,dl->el) = 1. If not such an structure
		 is found, *PDBL is inserted at the end of the list *dl
		 except if there is a *l such
		 that test_el(el,dl->el) returns a value 
		 different than 0 or 1 (see below).

 * ARGUMENTS:    PDBL *dl : list 
 *               PDBL *PDBL : structure to be inserted
 *               long (*test_el) () :

 * ALGORITHM:    inserts the PDBL pointer *PDBL right before
                 a PDBL pointer *l in the list *dl such that 
		 test_el(PDBL->el,l->el) = 1.If not such an structure
		 is found, *PDBL is inserted at the end of the list *dl
		 except if there is a *l such
		 that test_el(PDBL->el,dl->el) returns a value 
		 different than 0 or 1.
		 
		 If you design a test_el function
		 returning other values than 0 or 1 in some cases,
		 insere_PDBL will exit without inserting *PDBL
		 when encountering such cases. In this case, test_PDBL 
		 returns *dl directly.

		 In any other case, *PDBL will be inserted.
		 If *dl is null, returns *PDBL.

		 If l == dl => PDBL becomes the head of the list
		 and the function returns *PDBL.

 * RETURN:       PDBL *
 * SIDE EFFECT:  
 * CALLS:    dl_member()    
 * CALLED BY:    
 ***************/

PDBL *insere_dbl(PDBL *dl,PDBL *dbl,long (*test_el)())
     

{

    PDBL *l;

    if(dl == NULL){
	dbl->last = dbl->next = NULL;
	return(dbl);

    }

    l = dl_member(dl,dbl->el,test_el);

    if(l == &TEST_PDBL){ /* not found */

	/* l = end of the list */
        /* inserting at the end of the list */

	if(l->el != NULL){
	    l = dlbins(dbl,l->el);
	    return(dl);
	}

	/* 

	   Returning dl directly.
	   case where there was an element l->el such that
	   test_el(PDBL->el,l->el) 
	   returns a non null value different from 1. 
	   Since it is up to the user to decide what 
	   test_el returns in any case, it can be used in case  
	   of equality when ordering
	   for example.

	 */

	return(dl);
    }

    if( l == dl){

	l = dlhins(dbl,l);

	return(l);

    }

    l = dlhins(dbl,l);

    return(dl);

}


/***************
 * FUNCTION:     insere_PDBL
 * PURPOSE:      creates a PDBL structure and 
                 inserts the element *el right before
                 a PDBL pointer in the list *dl such that 
		 test_el(el,dl->el) = 1. 
		 If not such an structure
		 is found, *el is inserted at the end 
		 of the list *dl except if there is a *l such
		 that test_el(el,dl->el) returns a value 
		 different than 0 or 1 (see below).
		 
 * ARGUMENTS:    PDBL *dl : list 
 *               void *el : element to be inserted
 *               long (*test_el) () :

 * ALGORITHM:    creates a PDBL structure *la and put *el in it.
                 inserts the PDBL pointer *la right before
                 a PDBL pointer *l in the list *dl such that 
		 test_el(el,l->el) = 1. If not such an structure
		 is found, *la is inserted at the end of the list *dl
		 except if there is a *l such
		 that test_el(el,dl->el) returns a value 
		 different than 0 or 1.
		 
		 If you design a test_el function
		 returning other values than 0 or 1 in some cases,
		 insere_PDBL will exit without inserting *la
		 when encountering such cases. In this case, test_PDBL 
		 returns *dl directly.

		 In any other case, *la will be inserted.
		 If *dl is null, returns *la.

		 If l == dl => la becomes the head of the list
		 and the function returns *la.

 * RETURN:       PDBL *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

PDBL *insere_el(PDBL *dl,void *el,long (* test_el)())
     
{

    PDBL *l,*la;
    
    la = (PDBL *)malloc(sizeof(PDBL));
    la->el = el;

    if(dl == NULL){
	la->last = la->next = NULL;
	return(la);

    }

    l = dl_member(dl,el,test_el);

    if(l == &TEST_PDBL){
	if(l->el != NULL){
	    l = dlbins(la,l->el);
	    return(dl);
	}
	/* 
	   freeing *la and 
	   returning dl directly.
	   case where there was an element l->el such that
	   test_el(el,l->el) 
	   returns a non null value different from 1. 
	   Since it is up to the user to decide what 
	   test_el returns in any case, it can be used in case  
	   of equality when ordering
	   for example.

	 */
	
	free(la);
	return(dl);
	
    }

    if( l == dl){

	l = dlhins(la,l);

	return(l);

    }

    l = dlhins(la,l);

    return(dl);

}


/***************
 * FUNCTION:     dl_add_elm
 * PURPOSE:      adds the element "el" only if it doesn't belong
                 already to "dl"
 * ARGUMENTS:    PDBL *dl :
 *               void *el :
 * ALGORITHM:    uses the search function dlelm(). If "el" not found
                 does dlhcons() as above
 * RETURN:       PDBL *: returns the new header of the list
 * SIDE EFFECT:  
 * CALLS:        dlelm(), dlhcons()
 * CALLED BY:    
 ***************/

PDBL *dl_add_elm(PDBL *dl,void *el)


{

    PDBL *l;

    if((l = dlelm(dl,el)) == NULL)
	l = dlhcons(dl,el);

    return(l);

}


/***************
 * FUNCTION:     dlhins
 * PURPOSE:      inserts a PDBL structure in a list
 * ARGUMENTS:    PDBL *l :
 *               PDBL *dl :
 * ALGORITHM:    inserts "l" right before "dl"
                 "l" is supposed already allocated.
 * RETURN:       PDBL *: returns "l" itself
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


PDBL *dlhins(PDBL *l,PDBL *dl)

{

    if(dl != NULL){
	if(dl->last != NULL)
	    dl->last->next = l;

	l->last = dl->last;
	dl->last = l;
	l->next = dl;

    }

    else
	l->next = l->last = NULL;
    return(l);

}

/***************
 * FUNCTION:     dlbins
 * PURPOSE:      inserts a PDBL structure in a list
 * ARGUMENTS:    PDBL *l :
 *               PDBL *dl :
 * ALGORITHM:    inserts "l" right after "dl"
                 "l" is supposed already allocated.
 * RETURN:       PDBL * : returns "l" itself
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

PDBL *dlbins(PDBL *l,PDBL *dl)
    

{

    if(dl != NULL){
	if(dl->next != NULL)
	    dl->next->last = l;
	l->next = dl->next;
	dl->next = l;
	l->last = dl;
	return(dl);
    }

    l->last = l->next = NULL;
    return(l);

}

/***************************************************
  Searching in a list
***************************************************/


/***************
 * FUNCTION:     dlfin
 * PURPOSE:      looks for the end of a list.
 * ARGUMENTS:    PDBL *l :
 * ALGORITHM:    Looks for the last PDBL structure of
                 the list. Returns NULL if the list is
		 empty or if the list is closed.
 * RETURN:       PDBL * : last PDBL struture of the list, if any.
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


PDBL *dlfin(PDBL *l)
    
{

    PDBL *la;

    if(l == NULL)
	return(NULL);

    la = l;

    while(dls(la) != NULL){
	if(dls(la) == l){

	    return(NULL);
	}

	la = dls(la);
    }

    return(la);
}


/***************
 * FUNCTION:     dldeb
 * PURPOSE:      looks for the begining of a list.
 * ARGUMENTS:    PDBL *l :
 * ALGORITHM:    Looks for the first PDBL structure of
                 the list. Returns NULL if the list is
		 empty or if the list is closed.
 * RETURN:       PDBL * : first PDBL struture of the list, if any.
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/
PDBL *dldeb(PDBL *l)
     
{

    PDBL *la;

    if(l == NULL)
	return(NULL);

    la = l;

    while(dlp(la) != NULL){
	if(dlp(la) == l){

	    return(NULL);
	}

	la = dlp(la);
    }

    return(la);
}


/***************
 * FUNCTION:     dlelm
 * PURPOSE:      looks for an element in a list.
 * ARGUMENTS:    PDBL *l :
 *               void *el :
 * ALGORITHM:    Looks for the element "el" in the list "l".
                 Returns NULL if "el" is not found.
		 Returns the PDBL structure containing "el" if any.
		 Does not assume "l" is the begining of the list.
		 Looks forward first. Then, if "el" still not found,
		 and the list is not closed,  looks backwards.
		 Returns the first occurence of "el" only.
 * RETURN:       PDBL *: pointer on the PDBL structure 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


PDBL *dlelm(PDBL *l,void *el)
{

    PDBL *la;

    if(l == NULL)
	return(NULL);
    /* if el in first structure */
    if(dlel(l) == el)
	return(l);

    la = dls(l);

    /* looking till the end of the list, if any ...*/
    while( la != NULL){

	if(dlel(la) == el)
	    return(la);

	/* the list is closed */
	if(la == l  ){
	    /* element not found */
	    return(NULL);
	}
	la = dls(la);

    }

    la = dlp(l);

    while( la != NULL){

	if(dlel(la) == el)
	    return(la);

	if(la == l ){
	    printf(" dlelm() : WARNING  list closed only backwards !! \n");
	    return(NULL);
	}

	la = dlp(la);
    }

    return(NULL);

}




/***************
 * FUNCTION:     dlelm_lim_av
 * PURPOSE:      looks for the element "el" in "dl" 
                 going forward only and starting from "l"
 * ARGUMENTS:    PDBL *dl : header of the list
 *               PDBL *l : starting point
 *               void *el : element to be found
 * ALGORITHM:    Looks for the element "el" in the list "dl" starting
                 from "l". 
		 Returns NULL if "el" is not found.
		 Returns the PDBL structure containing "el" if any.
		 Does not assume "l" is the begining of the list.
		 Looks forward only.
		 Stops if either the list is finished or if it arrives
		 at "dl". Corrollary: "dl" does not have any effect if it 
		 is before "l" and "l" is an open list.
		 WARNING: if "l" does not belong to the list pointed to 
		 by "dl" and if "l" points to a closed list, this will go
		 forever. 
		 
		 Returns the first occurence of "el" only.
 * RETURN:       PDBL *: first structure containing "el" or "NULL"
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

PDBL *dlelm_lim_av(PDBL *dl,PDBL *l,void *el)
  
{

    PDBL *la;

    if(l == NULL)
	return(NULL);
    if(dlel(l) == el)
	return(l);

    la = dls(l);

    while( la != NULL){

	if(dlel(la) == el)
	    return(la);

	if(la == dl  ){
	    return(NULL);
	}
	la = dls(la);

    }
    return(NULL);

}

/***************
 * FUNCTION:     dlelm_lim_av
 * PURPOSE:      looks for the element "el" in "dl" 
                 going backwards only and starting from "l"
 * ARGUMENTS:    PDBL *dl : header of the list
 *               PDBL *l : starting point
 *               void *el : element to be found
 * ALGORITHM:    Looks for the element "el" in the list "dl" starting
                 from "l". 
		 Returns NULL if "el" is not found.
		 Returns the PDBL structure containing "el" if any.
		 Does not assume "l" is the begining of the list.
		 Looks backwards only.
		 Stops if either the list is finished or if it arrives
		 at "dl". Corrollary: "dl" does not have any effect if it 
		 is after "l" and "l" is an open list.
		 WARNING: if "l" does not belong to the list pointed to 
		 by "dl" and if "l" points to a closed list, this will go
		 forever. 
		 
		 Returns the first occurence of "el" only.
 * RETURN:       PDBL *: first structure containing "el" or "NULL"
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

PDBL *dlelm_lim_ar(PDBL *dl,PDBL *l,void *el)
    
{

    PDBL *la;

    if(l == NULL)
	return(NULL);
    if(dlel(l) == el)
	return(l);

    la = dlp(l);

    while( la != NULL){

	if(dlel(la) == el)
	    return(la);

	if(la == dl  ){
	    return(NULL);
	}
	la = dlp(la);

    }
    return(NULL);

}


/***************
 * FUNCTION:     dl_test_inter
 * PURPOSE:      test if there is an intersection, in term 
                 of element, between l1 and l2.
 * ARGUMENTS:    PDBL *l1 :
 *               PDBL *l2 :
 * ALGORITHM:    For each element el in the  list l1, applies the function
                 dlelm(l2,el). 
		 Doesn't do backsearch on l1 => if l1 is open, only elements
		 after l1 will be checked.
		 Returns as soon an intersection is found.
 * RETURN:       long : 0 is no intersection 
                        1 if at least 1 intersection
 * SIDE EFFECT:  
 * CALLS:        dlelm().
 * CALLED BY:    
 ***************/

long dl_test_inter(PDBL *l1,PDBL *l2)
     

{

  PDBL *la;
  if(l1 == NULL || l2 == NULL)
    return(0);

  la = l1;

  do{
    if(dlelm(l2,dlel(la)) != NULL)
      return(1);

    la = dls(la);
  }while(la != NULL && la != l1);

  return(0);

}



/***************
 * FUNCTION:     dl_member
 * PURPOSE:      looks for a PDBL struture ls in "l" containing an element
                 dlel(ls) satisfying member(el,dlel(ls)) != 0.
                 Member is a function defined by the user. 
		 Useful for ordering a list using any order relationship.
  * ARGUMENTS:    PDBL *l :
 *               void *el :
 *               long (*member) () :
 * ALGORITHM:    for each ls in l computes 
                      test = member(el,dlel(ls))
                 if test  != 0
		    if test = 1
                       returns ls
                    else if test != 1
		       returns TEST_PDBL with:
                         TEST_PDBL.last = ls
                         TEST_PDBL.el = NULL
                    endif
                 endif
                 
		 if no structure found:
                    if closed list 
                      returns TEST_PDBL with
                         TEST_PDBL.last = NULL
			 TEST_PDBL.el = l->last
                    else if open list
                      returns TEST_PDBL with
                         TEST_PDBL.last = NULL
			 TEST_PDBL.el = pointer to the last struture of
                                   the list
	         Returns NULL if l is empty or if l is badly linked.

 * RETURN:       PDBL *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

PDBL *dl_member(PDBL *l,void *el,long (*member)())
    
{            

    PDBL *la,*li;
    long val;

    if(l == NULL)
	return(NULL);
    if((val = member(el,dlel(l))) == 1 )

	return(l);

    if( val != 0){
	TEST_PDBL.el = NULL;
	TEST_PDBL.last = l;
	return(&TEST_PDBL);
    }
    li = l;
    la = dls(l);

    while( la != NULL){

	if((val = member(el,dlel(la))) == 1)
	    return(la);

	if( val != 0){
	    TEST_PDBL.el = NULL;
	    TEST_PDBL.last = la;
	    return(&TEST_PDBL);
	}
	if(la == l){

	    /* closed list, no element found */

	    TEST_PDBL.el = (char *)(l->last);
	    TEST_PDBL.last = NULL;
	    return(&TEST_PDBL);
	}

	li = la;
	if((la =  dls(la)) == NULL)

	    break; /* end of list */



    }

    /* at this point the last non null struture of the list is in "li" */

    /* looking backwards */

    if( (la = dlp(l)) == NULL){

	/* element not in the list */

	TEST_PDBL.el = (char *)(li);
	TEST_PDBL.last = NULL;
	return(&TEST_PDBL);

    }

    

    while(la != NULL){


	if((val = member(el,dlel(la))) == 1)
	    return(la);

	if( val != 0){
	    TEST_PDBL.el = NULL;
	    TEST_PDBL.last = la;
	    return(&TEST_PDBL);
	}

	if(la == l ){

	    /* badly linked list (closed only backwards)  */

	    printf(" dlelm() : list closed only backwards  !!! \n");

	    return(NULL);

	}

	la = dlp(la);
      }

    
    /* element not in the list*/
    TEST_PDBL.last = NULL;
    TEST_PDBL.el = (char *)li ;
    return(&TEST_PDBL);
    
    
    
  }


/***************
 * FUNCTION:     dl_member_lim_av
 * PURPOSE:      looks for a PDBL struture ls in "dl" containing an element
                 dlel(ls) satisfying member(el,dlel(ls)) != 0.
		 Looks forward only, starting from "l".

                 Member is a function defined by the user. 
		 Useful for ordering a list using using 
		 a user defined order relationship.
  * ARGUMENTS:   PDBL *dl; 
                 PDBL *l :
 *               void *el :
 *               long (*member) () :

 * ALGORITHM:    for each ls in l computes 
                      test = member(el,dlel(ls))
                 if test  != 0
		    if test = 1
                       returns ls
                    else if test != 1
		       returns TEST_PDBL with:
                         TEST_PDBL.last = ls
                         TEST_PDBL.el = NULL
                    endif
                 endif
                 
		 if no structure found:
                    if closed list 
                      returns TEST_PDBL with
                         TEST_PDBL.last = NULL
			 TEST_PDBL.el = l->last
                    else if open list
                      returns TEST_PDBL with
                         TEST_PDBL.last = NULL
			 TEST_PDBL.el = pointer to the last struture of
                                   the list
	         Returns NULL if l is empty or if l is badly linked.

 * RETURN:       PDBL *
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

PDBL *dl_member_lim_av(PDBL *dl,PDBL *l,void *el,long (*member)())
    
{

    PDBL *la,*li;
    long val;

    if(l == NULL)
	return(NULL);
    if((val = member(el,dlel(l))) == 1 )

	return(l);

    if( val != 0){
      TEST_PDBL.last = l;
      TEST_PDBL.el = NULL;
      return(&TEST_PDBL);
    }
    li = l;
    la = dls(l);

    while( la != NULL){

	if((val = member(el,dlel(la))) == 1)
	    return(la);

	if( val != 0){
	  TEST_PDBL.last = la;
	  TEST_PDBL.el = NULL;
	  return(&TEST_PDBL);
	}
	if(la == dl){

	    /* closed list */
	  TEST_PDBL.last = NULL;
	  TEST_PDBL.el = (char *)(dl->last);
	  return(&TEST_PDBL);
	}

	li = la;
	la =  dls(la);

      }

    /* element not in the list */
    TEST_PDBL.last = NULL;
    TEST_PDBL.el = (char *)(li);
    return(&TEST_PDBL);

}

/***************************************************************************
  Destruction functions
  **************************************************************************/


/***************
 * FUNCTION:     dldel
 * PURPOSE:      takes the PDBL struture "lt" off the list "l" and 
                 frees the memory corresponding to "lt".
 * ARGUMENTS:    PDBL *l :
 *               PDBL *lt :
 * ALGORITHM:    takes the PDBL structure "lt" off the list "l" and 
                 frees the memory corresponding to "lt"
		 DOES NOT FREE the element "lt->el".
		 returns "l" if l != lt
                 if l = lt 
                   returns l->next if l->next != NULL
                 else
                   returns l->last
 * RETURNS       PDBL *:
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


PDBL *dldel(PDBL *l,PDBL *lt)
  
{

    PDBL *la;

    if(lt == NULL || l == NULL)
	return(l);

    if(l == lt){
	if( l->next == NULL)
	    la = l->last;
	else if( l->next != l)
	    la = l->next;
	else {
	    free(lt);
	    
	    return(NULL);
	}

    }
    else
	la = l;

    if( lt->next != NULL )
	lt->next->last = lt->last;
    if( lt->last != NULL)
	lt->last->next = lt->next;

    free(lt);
    
    return(la);

}




/***************
 * FUNCTION:     dbdel_el
 * PURPOSE:      Frees the element contained by "lt"
                 using the user function "lib_el()"
                 takes the PDBL struture "lt" off the list "l" and 
                 frees the memory corresponding to "lt".
 * ARGUMENTS:    PDBL *l :
 *               PDBL *lt :
 *               void (*lib_el):
 * ALGORITHM:    
                 takes the PDBL structure "lt" off the list "l" and 
                 frees the memory corresponding to "lt"
		 Frees the element contained by "lt"
                 using the user function "lib_el()"
		 returns "l" if l != lt
                 if l = lt 
                   returns l->next if l->next != NULL
                 else
                   returns l->last
 * RETURNS       PDBL *:
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


PDBL *dbdel_el(PDBL *l,PDBL *lt,void (*lib_el)())
     
{

    PDBL *la;

    lib_el(dlel(lt));

    la = dldel(l,lt);

    return(la);

}


/***************
 * FUNCTION:     dldel_s
 * PURPOSE:      takes the PDBL struture "lt" off the list "l" and 
                 frees the memory corresponding to "lt".
		 Returns lt->next execpt if lt or l are NULL.
 * ARGUMENTS:    PDBL *l :
 *               PDBL *lt :
 * ALGORITHM:    takes the PDBL structure "lt" off the list "l" and 
                 frees the memory corresponding to "lt"
		 DOES NOT FREE the element "lt->el".
		 returns NULL if lt or l are NULL
                 otherwise, return always lt->next.
                 
 * RETURNS       PDBL *:
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


PDBL *dldel_s(PDBL *l,PDBL *lt)
  
{

    PDBL *la;

    if(lt == NULL || l == NULL)
	return(l);

    la = lt->next;

    if(lt->last != NULL)
	lt->last->next = la;
    if(la != NULL)
	la->last = lt->last;
    
    free(lt);

    return(la);

}

/***************
 * FUNCTION:     dldel_p
 * PURPOSE:      takes the PDBL structure "lt" off the list "l" and 
                 frees the memory corresponding to "lt".
		 Returns lt->last except if lt or l are NULL.
 * ARGUMENTS:    PDBL *l :
 *               PDBL *lt :
 * ALGORITHM:    takes the PDBL structure "lt" off the list "l" and 
                 frees the memory corresponding to "lt"
		 DOES NOT FREE the element "lt->el".
		 returns NULL if lt or l are NULL
                 otherwise, return always lt->last.
                 
 * RETURNS       PDBL *:
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

PDBL *dldel_p(PDBL *l,PDBL *lt)
    

{

    PDBL *la;

    if(lt == NULL || l == NULL)
	return(l);

    la = lt->last;

    if(lt->next != NULL)
	lt->next->last = la;
    if(la != NULL)
	la->next = lt->next;

    free(lt);
    
    return(la);

}


/***************
 * FUNCTION:     dl_del_elm
 * PURPOSE:      removes the element "el" from the list "dl" if
                 "el" belongs to "dl".
		 Does not free "el" itself.
 * ARGUMENTS:    PDBL *dl :
 *               void *el :
 * ALGORITHM:    grab the pointer to the PDBL structure containing "el"
                 and removes it, if any.
		 Removes only 1 instance of "el" in the list.
 * RETURN:       PDBL *: new header of the list in case the previous one
                        was removed in the process.
 * SIDE EFFECT:  
 * CALLS:        dlelm(), dldel()
 * CALLED BY:    
 ***************/

PDBL *dl_del_elm(PDBL *dl,void *el)
    

{

    PDBL *l;

    if(( l = dlelm(dl,el)) != NULL)
	l = dldel(dl,l);

    return(l);

}




/***************
 * FUNCTION:     dl_sub
 * PURPOSE:      takes the PDBL structure "lt" off the list "l" 
                 does not free the memory corresponding to "lt".
 * ARGUMENTS:    PDBL *l :
 *               PDBL *lt :
 * ALGORITHM:    takes off the PDBL structure "lt" from the list "l" 
                 DOES NOT FREE the memory corresponding to "lt"
		 DOES NOT FREE the element "lt->el".
		 returns "l" if l != lt
                 if l = lt 
                   returns l->next if l->next != NULL
                 else
                   returns l->last
 * RETURNS       PDBL *:
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

PDBL *dl_sub(PDBL *l,PDBL *lt)
    
{

    PDBL *la;

    if(lt == NULL || l == NULL)
	return(l);

    if(l == lt){
	if( l->next == NULL)
	    la = l->last;
	else if( l->next != l)
	    la = l->next;
	else {
	    return(NULL);
	}

    }
    else
	la = l;

    if( lt->next != NULL )
	lt->next->last = lt->last;
    if( lt->last != NULL)
	lt->last->next = lt->next;

    return(la);

}


/***************
 * FUNCTION:     dl_dep_av
 * PURPOSE:      removes the PDBL structure "lt" from its list
                 and puts it  before "l" 
 * ARGUMENTS:    PDBL *l :
 *               PDBL *lt :
 * ALGORITHM:    removes the structure "lt" from the list "lt". 
                 inserts lt before l
		 returns the new list "lt" (see dl_sub())
		 NOTE: l and lt may not belong to the same list before this call.
 * RETURN:       PDBL *
 * SIDE EFFECT:  
 * CALLS:        dl_sub(), dlhins()
 * CALLED BY:    
 ***************/



PDBL *dl_dep_av(PDBL *l,PDBL *lt)
    
{
    PDBL *la;

    la = dl_sub(lt,lt);
    dlhins(lt,l);
    return(la);
}


/***************
 * FUNCTION:     dl_dep_ap
 * PURPOSE:      removes the PDBL structure "lt" from its list
                 and puts it  after "l" 
 * ARGUMENTS:    PDBL *l :
 *               PDBL *lt :
 * ALGORITHM:    removes the structure "lt" from the list "lt". 
                 inserts lt after l
		 returns the new list "lt" (see dl_sub())
		 NOTE: l and lt may not belong to the same list before this call.  
 * RETURN:       PDBL *
 * SIDE EFFECT:  
 * CALLS:        dl_sub(), dlhins()
 * CALLED BY:    
 ***************/


PDBL *dl_dep_ap(PDBL *l,PDBL *lt)

{
    PDBL *la;

    la = dl_sub(lt,lt);
    dlbins(lt,l);
    return(la);
}



/***************
 * FUNCTION:     dlib_list
 * PURPOSE:      frees all PDBL structures in the list l (goes forwards 
                 and backwards).
		 DOES NOT FREE the elements pointed to by the list.
 * ARGUMENTS:    PDBL *l :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void dlib_list(PDBL *l)
   

{
    PDBL *dl,*la,*li;

    if(l == NULL)
	return;

    la = l;

    li = l->last;

    /* destruction en marche avant */

    do
	{
	    dl = la->next;
	    free(la);
	    
	    la = dl ;

	} while(dl != NULL && dl != l);

    /* case of an open list: going backwards */

    if( dl == NULL){
	if(li != NULL){
	    do {
		dl = li->last;
		free(li);
	    
		li = dl;

	    }while(dl != NULL);
	}
    }
}




/***************
 * FUNCTION:     dlib_list_el
 * PURPOSE:      frees both the PDBL structures of the list "l" and the elements
                 pointed to by the list. 
		 Uses the user function parameter lib_el() for freeing the internal 
		 elements.
 * ARGUMENTS:    PDBL *l :
 *               void (*lib_el) () :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void dlib_list_el(PDBL *l,void (*lib_el)())
     
{
    PDBL *dl,*la,*li;

    if(l == NULL)
	return;
    
    la = l;

    li = l->last;

    /* going forward */

    do
	{
	    dl = la->next;
	    lib_el(la->el);
	    free(la);
	    
	    la = dl ;

	} while(dl != NULL || dl == l);

    /* case of an open list */

    if( dl == NULL){

	/* l wasn't the header */

	if(li != NULL){

	    /* going backwards */

	    do {

		dl = li->last;
		lib_el(li->el);
		free(li);
		
		li = dl;

	    }while(dl != NULL);


	}
    }
}


/**********************************************************************
  Opening and closing a list
  **********************************************************************/




/***************
 * FUNCTION:     dl_close
 * PURPOSE:      close the list pointed to by "dl"
 * ARGUMENTS:    PDBL *dl :
 * ALGORITHM:    takes any pointer on a  PDBL structure and close the list
                 attached to it. 
		 uses dldeb() and dlfin()
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        dldeb(), dlfin()
 * CALLED BY:    
 ***************/

void dl_close(PDBL *dl)
    
{

    PDBL *ddeb,*dfin;

    ddeb = dldeb(dl);
    dfin = dlfin(dl);
   
    ddeb->last = dfin;
    dfin->next = ddeb;

}


/***************
 * FUNCTION:     dl_open_a
 * PURPOSE:      opens the list before the pointer "dl"
 * ARGUMENTS:    PDBL *dl :
 * ALGORITHM:    Looks if there is a pointer before "dl".
                 if yes, deconnects "dl" from its previous element.
		 Warning: doesn't do an explicit test to see if the list
		 is closed. If the list is not closed, "dl" will be broken in two 
		 pieces, one of which will be lost.
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void dl_open_a(PDBL *dl)
     

{

    if(dlp(dl) != NULL)
	dls(dlp(dl)) = NULL;

    dlp(dl) = NULL;

}


/***************
 * FUNCTION:     dl_open_b
 * PURPOSE:      opens the list after the pointer "dl"
 * ARGUMENTS:    PDBL *dl :
 * ALGORITHM:    Looks if there is a pointer after "dl".
                 if yes, deconnects "dl" from its next structure.
		 Warning: doesn't do an explicit test to see if the list
		 is closed. If the list is not closed, "dl" will be broken in two 
		 pieces, one of which will be lost.
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void dl_open_b(PDBL *dl)
     

{

    if(dls(dl) != NULL)
	dlp(dls(dl)) = NULL;

    dls(dl) = NULL;

}





/***************
 * FUNCTION:     dl_colle
 * PURPOSE:      sticks l2 after l1. No duplication.
 * ARGUMENTS:    PDBL *l1 :
 *               PDBL *l2 :
 * ALGORITHM:    Looks for the end of l1 and hook l2 to it.
                 Warning : there is no test about  l2 being open !
		 At the end, l2 belongs to the list l1.
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

void dl_colle(PDBL *l1,PDBL *l2)


{
    PDBL *l;

    if(l1 == NULL || l2 == NULL)
	return;

    if ((l = dlfin(l1)) == NULL)
    {
      printf("list l1 is closed. Can't stick lists together\n");
      return;
    }

    l->next = l2;
    l2->last = l;

}


/***************
 * FUNCTION:     dl_copy
 * PURPOSE:      copies l1 onto l2. 
 * ARGUMENTS:    PDBL *l1 :
 *               PDBL *l2 :
 * ALGORITHM:    copies all elements included after l1 in l2.
                 returns the new l2 list.
		 It is a physical copy => l1 does not belong to l2 at the end.
		 If l1 is not closed, elements of l1 before l1 won't be copied.
 * RETURN:       PDBL *: new list l2
 * SIDE EFFECT:  
 * CALLS:        dlhcons()
 * CALLED BY:    
 ***************/

PDBL *dl_copy(PDBL *l1,PDBL *l2)
     

{

    PDBL *l,*lr;

    if( (l = l1) == NULL)
	return(l2);

    lr = l2;

    do{
	lr = dlhcons(lr,dlel(l));
	l = dls(l);

    }while(l != l1 && l != NULL);

    return(lr);

}

/************************
  Info functions
************************/


/***************
 * FUNCTION:     Db_length
 * PURPOSE:      returns the full length of the list "dl"
 * ARGUMENTS:    PDBL *dl :
 * ALGORITHM:    goes forward and backwards along the list.
                 Returns 0 if dl = NULL
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/
 

long Db_length(PDBL *dl)
     

{

    PDBL *l;
    long n = 0;

    if(dl == NULL)
	return(0);

    l = dl;

    do{

	n++;
	l = dls(l);
    }while(l != NULL && l != dl);

    /* list closed or dl = header of the list */
    if( l != NULL || dlp(dl) == NULL) 
      return(n);

    /* going backwards */
    l = dlp(dl);
    
    do{
	n++;
	l = dlp(l);
    }while(l != NULL);

    return(n);

}


/***************
 * FUNCTION:     Db_lg_av
 * PURPOSE:      gives the length of dl going forward only
 * ARGUMENTS:    PDBL dl :
 * ALGORITHM:    gives the length of dl going forward only.
                 Db_lg_av(dl) = Db_length(dl) => dl is closed
		 or dl is the header of the list.
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long Db_lg_av(PDBL *dl)
   

{

    PDBL *l;
    long n = 0;

    if(dl == NULL)
	return(0);

    l = dl;

    do{

	n++;
	l = dls(l);
    }while(l != NULL && l != dl);

    return(n);

}

/***************
 * FUNCTION:     Db_lg_ap
 * PURPOSE:      gives the length of dl going backward only
 * ARGUMENTS:    PDBL dl :
 * ALGORITHM:    gives the length of dl going backward only.
                 Db_lg_ap(dl) = Db_length(dl) => dl is closed
		 or dl is the header of the list.
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long Db_lg_ap(PDBL *dl)
    

{

    PDBL *l;
    long n = 0;

    if(dl == NULL)
	return(0);

    l = dl;

    do{

	n++;
	l = dlp(l);
    }while(l != NULL && l != dl);

    return(n);

}

/***************************
  renvoit 1 si on trouve l2
  apres l1. Corrolaire:
  si liste bouclee, renvoit
  toujours 1.
  le test Db_av(l,l)
  revient donc a tester si la
  liste est bouclee.
  ****************************/

/***************
 * FUNCTION:     PDBL_av
 * PURPOSE:      tests if l2 is after l1 in the list
                 NOTE: PDBL_av(l1,l1) tests if l1 is closed (at least forward).
 * ARGUMENTS:    PDBL *l1 :
 *               PDBL *l2 :
 * ALGORITHM:    
 * RETURN:       long 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long PDBL_av(PDBL *l1,PDBL *l2)
   
{
    PDBL *l;

    l = dls(l1);

    do{

	if( l == l2)
	    return(1);
	l = dls(l);

    }while( l != NULL && l != dls(l1));

    return(0);
}

/***************
 * FUNCTION:     PDBL_aff
 * PURPOSE:      Displays the contents of all  elements in 
                 the list pointed to by *PDBL using the user display
		 function "displ_func".
 * ARGUMENTS:    PDBL *PDBL :
 *               void *displ_func :
 * ALGORITHM:    
 * RETURN:       void 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


void PDBL_aff(PDBL *dbl,void (*displ_func)())
     

{

  PDBL *l;

  if(dbl == NULL)
    return;

  if((l = dldeb(dbl))== NULL)
     l = dbl;
  do{
    
    displ_func(dlel(l));
    l = dls(l);
  }while(l != NULL && l != dbl);


}











