

#include <math.h>

#ifdef vax
#define HUGE	1.701411733192644270e38
#endif
#ifdef mc68000
#ifndef HUGE
#define HUGE	1.79769313486231+308	/* IEEE infinity */
#endif
#define HUGEsp	3.40282346638529E+38	/* Largest binary float. */
#define SMALLsp	1.17549435082229E-38	/* Smallest (positive) binary float. */
#endif
#ifndef HUGEsp
#define HUGEsp 1.0e+38
#endif
#ifndef SMALLsp
#define SMALLsp 1.0e-45
#endif



#include "sge_prot.h"

long isamax(long n,double *sx, long incx )

/*
    PURPOSE
	Finds the index of element having max. absolute value.

    INPUT
	n	Number of elements to check.
	sx	Vector to be checked.
	incx	Every incx-th element is checked.

*/
{
    register double smax = 0.0e0;
    register int i, istmp = 0;

#ifndef abs
#define abs(x) ((x)<0.0?-(x):(x))
#endif
#ifdef alliant
  long isamax_();

  istmp = isamax_( &n, sx, &incx )-1;	/* Remember 0 is first. */
  return( istmp );
#else
  if( n <= 1 ) return( istmp );
  if( incx != 1 ) {
    /* Code for increment not equal to 1. */
    if( incx < 0 ) sx = sx + ((-n+1)*incx + 1);
    istmp = 0;
    smax  = abs( *sx );
    sx += incx;
    for( i=1; i<n; i++, sx+=incx )
      if( abs( *sx ) > smax ) {
	istmp = i;
	smax  = abs( *sx );
      }
    return( istmp );
  }
  /* Code for increment equal to 1. */
  istmp = 0;
  smax  = abs(*sx);
  sx++;
  for( i=1; i<n; i++, sx++ ) 
    if( abs( *sx ) > smax ) { 
      istmp = i;
      smax  = abs( *sx );
    }
  return( istmp );
#endif
}
 
double sasum(long n,double *sx,long incx )

/*
  PURPOSE
      Returns sum of magnitudes of single precision SX.
      sasum = sum from 0 to n-1 of  ABS(SX(1+I*INCX))

  INPUT
    n		Number of elements to multiply.
    sx		Pointer to double vector to take abs sum of.
    incx	Storage incrament for sx.

  RETURNS
    sasum       Double variable with the result.

  WARNINGS
    This routine uses the UN*X math library function fabs().

*/
{
#ifndef abs
#define abs(x) ((x)<0.0?-(x):(x)) 
#endif
    register double sum = 0.0;
    
    if( n<= 0 ) return( sum );
    if( incx == 1 ) {
	register int i, m;

        /* Code for increments equal to 1. */

        /* Clean-up loop so remaining vector length is a multiple of 6. */
	m = n % 6;
	if( m != 0 ) {
	    for( i=0; i<m; i++ )
	      sum += abs( sx[i] );
	if( n < 6 ) return( (double)sum );
	}
	for( i=m; i<n; i+=6 ) {
	    sum += abs( sx[i] ) + abs( sx[i+1]) + abs( sx[i+2] ) + 
	      abs( sx[i+3] ) + abs( sx[i+4] ) + abs( sx[i+5] );
	}
	return( (double)sum );
    } 
    else {
	register int i, ns;

        /* Code for increments not equal to 1. */
	ns = n*incx;
	for( i=0; i<ns; i+=incx ) {
	    sum += abs( sx[i] );
	}
	return( (double)sum );
    }
}

void saxpy(long n,double sa,double *sx,long incx,double *sy,long incy )

/*
  PURPOSE
    Vector times a scalar plus a vector.  sy = sy + sa*sx.

  INPUT
    n		Number of elements to multiply.
    sa		Scalar to multiply by (note that this is a double).
    sx		Pointer to double vector to scale.
    incx	Storage incrament for sx.
    sy		Pointer to double vector to add.
    incy	Storage incrament for sy.

  OUTPUT
    sy		sy = sy + sa*sx
*/
{
    register int i;
    register double ssa = (double)sa;

#ifdef alliant
    void saxpy_();
    double ssaa = (double)sa;
    
    saxpy_( &n, &ssaa, sx, &incx, sy, &incy );
    return;
#else
    
    if( n<=0 || ssa==0.0 ) return;
    if( incx == incy ) {
	if( incx == 1 ) {

	    /* Both increments = 1 */
	    for( i=0; i<n; i++ ) 
	      sy[i] += ssa*sx[i];
	    return;
	}
	if( incx>0 ) {

	    /* Equal, positive, non-unit increments. */
	    for( i=0; i<n; i++,sx+=incx,sy+=incx )
	      *sy += ssa*(*sx);
	    return;
	}
    }

    /* Unequal or negative increments. */
    if( incx < 0 ) sx += ((-n+1)*incx + 1);
    if( incy < 0 ) sy += ((-n+1)*incy + 1);
    for( i=0; i<n; i++,sx+=incx,sy+=incy ) 
      *sy += ssa*(*sx);
#endif
}
 
void saxpyx( long n,double sa,double *sx,long incx,double *sy,long incy )

/*
  PURPOSE
    Vector times a scalar plus a vector.  sx = sy + sa*sx.

  INPUT
    n		Number of elements to multiply.
    sa		Scalar to multiply by (note that this is a double).
    sx		Pointer to double vector to scale.
    incx	Storage incrament for sx.
    sy		Pointer to double vector to add.
    incy	Storage incrament for sy.

  OUTPUT
    sx		sx = sy + sa*sx
*/
{
    register i;
    register double ssa = (double)sa;

    if( n<=0 || ssa==0.0 ) return;
    if( incx == incy ) {
	if( incx == 1 ) {

	    /* Both increments = 1 */
	    for( i=0; i<n; i++ ) 
	      sx[i] = sy[i] + ssa*sx[i];
	    return;
	}
	if( incx>0 ) {

	    /* Equal, positive, non-unit increments. */
	    for( i=0; i<n; i++, sx+=incx, sy+=incx)
	      *sx = *sy + ssa*(*sx);
	    return;
	}
    }

    /* Unequal or negative increments. */
    if( incx < 0 ) sx += ((-n+1)*incx + 1);
    if( incy < 0 ) sy += ((-n+1)*incy + 1);
    for( i=0; i<n; i++,sx+=incx,sy+=incy ) 
      *sx = *sy + ssa*(*sx);
}

void scopy(long n,double *sx,long incx,double *sy,long incy )

/*
    PURPOSE
        Copies vector sx into vector sy.
 
    INPUT
        n    Number of components to copy.
	sx   Source vector
	incx Index increment for sx.
        incy Index increment for sy.
 
    OUTPUT
        sy   Destination vector.
*/
{
    register int i;

#ifdef alliant
    void scopy_();

    scopy_( &n, sx, &incx, sy, &incy );
    return;
#else

    if( n<1  ) return;
    if( incx == incy ) {
	if( incx == 1 ) {

	    /* Both increments = 1 */
	    for( i=0; i<n; i++ )
	      sy[i] = sx[i];
	    return;
	}
	if( incx > 0 ) {

	    /* Equal, positive, non-unit increments. */
	    for( i=0; i<n; i++, sx+=incx, sy+=incx)
	      *sy = *sx;
	    return;
	}
    }

    /* Non-equal or negative increments. */
    if( incx < 0 ) sx += ((-n+1)*incx + 1);
    if( incy < 0 ) sy += ((-n+1)*incy + 1);
    for( i=0; i<n; i++,sx+=incx,sy+=incy ) 
      (*sx) = (*sy);
    return;
#endif
}

double sdot(long n,double *sx,long incx,double *sy,long incy )

/*
    PURPOSE
        Forms the dot product of a vector.

    INPUT
        n       Number of elements to sum.
        sx      Address of first element of x vector.
        incx    Incrament for the x vector.
        sy      Address of first element of y vector.
        incy    incrament for the y vector.

    OUPUT
        sdot    Dot product x and y.  Double returned
		due to `C' language features.
*/
{
    register int i;
    register double stemp = 0.0e0;

    if( n<1 ) return( (double)stemp );
    if( incx == incy ) {
	if( incx == 1 ) {

	    /* Both increments = 1 */
	    for( i=0; i<n; i++ )
	      stemp += sx[i]*sy[i];
	    return( (double)stemp );
	}
	if( incx>0 ) {

	    /* Equal, positive, non-unit increments. */
	    for( i=0; i<n; i++, sx+=incx, sy+=incx)
	      stemp += (*sx)*(*sy);
	    return( (double)stemp );
	}
    }

    /* Unequal or negative increments. */
    if( incx < 0 ) sx += ((-n+1)*incx + 1);
    if( incy < 0 ) sy += ((-n+1)*incy + 1);
    for( i=0; i<n; i++,sx+=incx,sy+=incy ) 
      stemp += (*sx)*(*sy);
    return( (double)stemp );
}				/* End of ---SDOT--- */

double snrm2(long n,double *sx,long incx )

/*
    PURPOSE
        Computes the Euclidean norm of sx while being
	very careful of distructive underflow and overflow.

    INPUT
        n       Number of elements to use.
        sx      Address of first element of x vector.
        incx    Incrament for the x vector (>0).

    OUPUT
        snrm2   Euclidean norm of sx.  Returns double
		due to `C' language features.
    REMARKS
        This algorithm proceeds in four steps.
	1) scan zero components.
	2) do phase 2 when component is near underflow.
*/
{
  register int i;
  static double cutlo, cuthi;
  double sum = 0.0e0, hitst, r1mach();
  double xmax;

  if( n<1 || incx<1 ) return( sum );

  /* Calculate near underflow */
  if( cutlo == 0.0 ) cutlo = sqrt( SMALLsp/r1mach() );
  /* Calculate near  overflow */
  if( cuthi == 0.0 ) cuthi = sqrt( HUGEsp );
  hitst = cuthi/(double) n;
  i     = 0;

  /* Zero Sum. */
  while( *sx == 0.0 && i<n ) {
    i++;
    sx += incx;
  }
  if( i>=n ) return( sum );

START:
  if( abs( *sx ) > cutlo ) {
    for( ; i<n; i++, sx+=incx ) {		/* Loop over elements. */
      if( abs( *sx ) > hitst ) goto GOT_LARGE;  
      sum += (*sx) * (*sx);
    }
    sum = sqrt( sum );
    return( sum );				/* Sum completed normaly. */
  }
  else {					/* Small sum prepare for phase 2. */
    xmax  = abs( *sx );
    sx += incx;
    i++;
    sum += 1.0;
    for( ; i<n; i++, sx+=incx ) {
      if( abs( *sx ) > cutlo ) {		/* Got normal elem.  Rescale and process. */
	sum = (sum*xmax)*xmax;
	goto START;
      }
      if( abs( *sx ) > xmax ) {
	sum  = 1.0 + sum*(xmax /(*sx))*(xmax /(*sx));
	xmax = abs( *sx );
	continue;
      }
      sum += ((*sx)/xmax)*((*sx)/xmax);
    }
    return( (double)xmax*sqrt( sum ) );
  }						/* End of small sum. */

 GOT_LARGE:
  sum  = 1.0 + (sum/(*sx))/(*sx);		/* Rescale and process. */
  xmax = abs( *sx );
  sx   += incx;
  i++;
  for( ; i<n; i++, sx+=incx ) {
    if( abs( *sx ) > xmax ) {
      sum  = 1.0 + sum*(xmax /(*sx))*(xmax /(*sx));
      xmax = abs( *sx );
      continue;
    }
    sum += ((*sx)/xmax)*((*sx)/xmax);
  }
  return( (double)xmax*sqrt( sum ) );		/* End of small sum. */
}						/* End of ---SDOT--- */

double r1mach(void)
/***********************************************************************
 ****   This routine computes the unit roundoff for single precision 
 ****	of the machine.  This is defined as the smallest positive 
 ****	machine number u such that  1.0 + u .ne. 1.0
 ****	Returns a double due to `C' language features.
 **********************************************************************/
{
    double u = 1.0e0, comp;
 
    do {
        u *= 0.5e0;
        comp = 1.0e0 + u;
    }
    while( comp != 1.0e0 );
    return( (double)u*2.0e0 );
} /*-------------------- end of function r1mach ------------------------*/
 
long min0(long n,long a,long b,long c,long d,long e,long f,long g,long h,long i,
	  long j,long k,long l,long m,long o,long p )
/*
    PURPOSE
        Determine the minimum of the arguments a-p.
 
    INPUT
        n       Number of arguments to check 1 <= n <= 15.
        a-p     Integer arguments of which the minumum is desired.
 
    RETURNS
        min0    Minimum of a thru p.
*/

{
    long mt;
 
    if( n < 1 || n > 15 ) return( -1 );
    mt = a;
    if( n == 1 ) return( mt );
 
    if( mt > b ) mt = b;
    if( n == 2 ) return( mt );
 
    if( mt > c ) mt = c;
    if( n == 3 ) return( mt );
 
    if( mt > d ) mt = d;
    if( n == 4 ) return( mt );
 
    if( mt > e ) mt = e;
    if( n == 5 ) return( mt );
 
    if( mt > f ) mt = f;
    if( n == 6 ) return( mt );
 
    if( mt > g ) mt = g;
    if( n == 7 ) return( mt );
 
    if( mt > h ) mt = h;
    if( n == 8 ) return( mt );
 
    if( mt > i ) mt = i;
    if( n == 9 ) return( mt );
 
    if( mt > j  ) mt = j;
    if( n == 10 ) return( mt );
 
    if( mt > k  ) mt = k;
    if( n == 11 ) return( mt );
 
    if( mt > l  ) mt = l;
    if( n == 12 ) return( mt );
 
    if( mt > m ) mt = m;
    if( n == 13 ) return( mt );
 
    if( mt > o  ) mt = o;
    if( n == 14 ) return( mt );
 
    if( mt > p  ) mt = p;
    return( mt );
}
void sscal(long  n,double sa,double *sx,long incx )

/*
    PURPOSE
        Scales a vector by a constant.
 
    INPUT
        n    Number of components to scale.
        sa   Scale value (note that this is a double).
        sx   Vector to scale.
        incx Every incx-th element of sx will be scaled.
 
    OUTPUT
        sx   Scaled vector.
*/
{
    register double ssa = (double)sa;
    long i;

#ifdef alliant
    void sscal_();
    double ssaa = (double)sa;

    sscal_( &n, &ssaa, sx, &incx );
    return;
#else
 
    if( n < 1 ) return;

    /* Code for increment not equal to 1.*/
    if( incx != 1 ) {
	if( incx < 0 ) sx += (-n+1)*incx;
	for( i=0; i<n; i++, sx+=incx )
	  *sx *= ssa;
        return;
    }

    /*  Code for unit increment. */
    for( i=0; i<n; i++ ) 
      sx[i] *= ssa;
    return;
#endif
}
void vexopy(long n,double *v,double *x,double *y,long itype )

/*
  Purpose:
    To operate on the vectors x and y.

  Input:
    n		Number of elements to scale.
    x		First operand vector.
    y		Second operand vector.
    itype	Type of operation to perform:
		itype = 1 => '+'
		itype = 2 => '-'

  Output:
    v		Result vector of x op y.
*/
{
    register int i;

    if( n<1 ) return;

    if( itype == 1 )		/* ADDITION. */
      for( i=0; i<n; i++ )
	v[i] = x[i] + y[i];
    else			/* SUBTRACTION. */
      for( i=0; i<n; i++ )
	v[i] = x[i] - y[i];
}
void vfill(long n,double *v,double val )

/*
  Purpose
    To fill the DOUBLE vector v with the value val.
    Make sure that if you pass a value for val that you cast
    it to double, viz.,
		vfill( 100, vector, (double)0.0 );
*/
{
    register int i;
    register double vval = (double)val;

    if( n<1 ) return;
    for( i=0; i<n; i++ ) 
      v[i] = vval;
}
#ifdef COPYSIGN
double copysign(double x,double y )
/*
  PURPOSE
     This routine is recomended by the IEEE standard 754.  Most C 
     3M libraries contain this function.  If yours does not then
     compile this file with the -DCOPYSIGN option and the following
     code will be generated.  This routine copies the sign from y
     onto x.

  INPUT
     x		Thing to have it's sign changed.
     y		Variable to supply the sign.

  RETURNS
     copysign	Returns x with the sign of y (the fortran intrinsic
     		sign(x,y)).
*/
{
    /* Registers are used for reduced memory traffic. */
    register double xx = x;
    register double yy = y;

    if( xx*yy < 0.0 ) return( -xx );
    else return( xx );
}
#endif
