#ifdef VXWORKS
#include <vxWorks.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include "mat.h"
#include "sge_prot.h"
#include "matsge_p.h"




#define VERYSMALL(x) ( (x) > -1e-15 && (x) < 1e-15 )

#define Falloctemp(l,c) ( F_alloctemp ((l) * (c)) )
#define Salloctemp(l,c) ( F_alloctemp ((l) * ((c) + 1) / 2) )


/**********************************************************************

       Static Global variables for temporary computations
         
**********************************************************************/


double *F_alloctemp(long n);


/***************
 * FUNCTION:     mx_inv
 * PURPOSE:      inverting the f_matrix m1 into m2
 * ARGUMENTS:    double *m1 :f_matrix
 *               double *m2 :f_matrix
 *               long r :dimensions of m1 and m2
 *               long c :
 * ALGORITHM:    
 * RETURN:       long : = 0 if success
                        = 1 if bad conditionning of m1
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mx_inv (double *m1, double *m2,long r,long c)


{
  double detp;
  double temp;

  if (r != c)
    {
      printf ("mx_inv : non-square matrix\n");
      return (1);
    }

  switch (r)
    {
    case 1:
      *m2 = 1 / *m1;
      return (0);

    case 2:
      detp = m1[0] * m1 [3] - m1[1] * m1[2];
      if (VERYSMALL(detp))
	return(1);
      temp = m1[3];
      m2[3] = m1[0] / detp;
      m2[0] = temp / detp;
      m2[1] = - m1[1] / detp;
      m2[2] = - m1[2] / detp;
      return(0);
    default:
      return(mx_inv_sg(m1,m2,r,r));

    }

}


/********
  Solves the system m.x = b of r equations with c unknown with
  least square;

  dim(b) = r;
  dim(x) = c;

  x has to be allocated by the user
  r >= c

  returns 0 if ok

  returns 1 if bad conditionning
*********/

/***************
 * FUNCTION:     mx_lsqr
 * PURPOSE:      Solves the system m.x = b 
                 of r equations with c unknown 
                 using least square;
		 dim(b) = r;
		 dim(x) = c;

		 x has to be allocated by the user
		 r must be  greater than c

 * ARGUMENTS:    double *m : r by c f_matrix 
 *               double *x : c vector (output)
 *               double *b : r vector 
 *               long r : 
 *               long c :
 * ALGORITHM:    computes : x = (inverse(transpose(a).a)).transpose(a).b
 * RETURN:       long: returns 0 if ok
                       returns 1 if bad conditionning 
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/


long mx_lsqr(double *m,double *x,double *b,long r,long c)  
    
{
  double *Inv,*trm ;
  
  Inv = F_alloctemp(2*c*r);

  trm = Inv + c*r;
  
  mx_trmult(m,m,trm,r,c,r,c);

  if( mx_inv(trm,Inv,c,c) != 0)
    return(1);

  mx_multtr(Inv,m,trm,c,c,r,c);

  mx_mult(trm,b,x,c,r,r,1);

  return(0);

}


/***************
 * FUNCTION:     mxs_inv
 * PURPOSE:      inverting the s_matrix m1 into m2
 * ARGUMENTS:    double *m1 :s_matrix
 *               double *m2 :s_matrix
 *               long r :dimensions of m1 and m2
 *               long c :
 * ALGORITHM:    
 * RETURN:       long : = 0 if success
                        = 1 if bad conditionning of m1
 * SIDE EFFECT:  
 * CALLS:        
 * CALLED BY:    
 ***************/

long mxs_inv (double *m1, double *m2,long r,long c)

{

  double *aux;
  double detp;
  double temp;
  

  if (r != c)
    {
      printf ("mxs_inv : non-square matrix\n");
      return (1);
    }

  switch (r)
    {
    case 1:
      *m2 = 1 / *m1;
      return (0);
    case 2:
      detp = m1 [0] * m1 [2] - m1 [1] * m1 [1];
      if (VERYSMALL(detp))
	return(1);
      temp = m1[2];
      m2[2] = m1 [0] / detp;
      m2[0] = temp / detp;
      m2[1] = - m1 [1] / detp;
      return(0);
   default:
      aux = F_alloctemp(r*(r+1)/2);
      mxsf_copy(m1,aux,r,r);
      mx_inv_sg(aux,aux,r,r);
      mxfs_copy(aux,m2,r,r);
      return (0);
    }
}








